!!---------------------------------------------------------------------
!!.. Wave-equation-based inversion modules
!!.. OpenMP parallelized
!!.. Version 8.0
!!.. 1. This version includes the forward and adjoint operators
!!.. 2. This version also parallelizes the imaging condition
!!.. 3. This version uses the same sign for both source and receiver
!!..    wavefield and takes the conjugate of the source wavefield
!!..    after it is read from the disk immediately 
!!.. 4. This version includes complex image output/input (1/28/2009)
!!.. 5. This version supports output/input unstacked image 
!!..    (gather with surface parameters) (2/6/2009)
!!.. 6. This version fully passes dot-product, even with taper added
!!..    on (1/14/2010)
!!.. 7. This version supports output image on a coarser grid (1/14/2010)
!!.. 8. This version implements bulletproof implicit finite difference
!!..    and Fourier finite difference (8/2010)
!!.. 9. This version includes forward, adjoint and backward extrapolation
!!..    of wavefields, fully passes dot-product test for variable velocities 
!!..    with adjoint extrapolation (8/2010)
!!.. 10. This version implements phase error compensation filter for
!!..     for implicit finite difference (8/2010)
!!.. 11. This version preconditions the gradient with non-stationary
!!..     dip filters (10/2010)
!!---------------------------------------------------------------------
!!.. Author: Yaxun Tang, tang@sep.stanford.edu
!!---------------------------------------------------------------------
!!.. Note: 1. To make full usage of the threads, 
!!..          it is better to keep nws >= nth 
!!..       2. For threads safty purpose, the memory allocation is
!!..          considerably increased compared to serial implementation.
!!..          Mem of parallel ~ Mem of serial * nth
!!---------------------------------------------------------------------
module bwi_kernel_mod2
  use  base_blkcal_mod
  use  base_spline_mod
  use  bwi_struct_mod
  use  bwi_histos_mod
  use  bwi_glocal_mod
  use  bwi_fileio_mod
  use  bwi_velref_mod
  use  ani_wfdexp_mod
  use  bwi_wfdexp_mod
  use  bwi_imgcon_mod
  use  bwi_fftwv3_mod
  use  bwi_tapers_mod
  use  bwi_sbcond_mod
  use  ani_bornsc_mod
  use  bwi_rdmgen_mod
  use  bwi_hescpt_mod
  use  bwi_report_mod
  use  bwi_sbcond_mod
  use  ani_getcef_mod
  implicit none
  type(dat_grid)                              , private :: dgrid    !!.. receiver grid
  type(vel_grid)                              , private :: vgrid    !!.. velocity grid
  type(pro_grid)                              , private :: pgrid    !!.. propagation grid
  type(img_grid)                              , private :: igrid    !!.. image grid
  type(out_grid)                              , private :: ogrid    !!.. image output grid
  type(hes_grid)                              , private :: hgrid    !!.. Hessian grid
  type(cef_grid)                              , private :: cgrid    !!.. optimized coef grid

  type(block)                                 , private :: wblock

  type(bwi_ftag)                              , private :: ftag_

  real     , dimension(:,:)      , allocatable, private :: bvel     !!.. velocity model, backgorund
  real     , dimension(:,:)      , allocatable, private :: beta     !!.. eta      model, backgorund
  real     , dimension(:,:)      , allocatable, private :: bdel     !!.. delta    model, backgorund
  real     , dimension(:,:,:)    , allocatable, private :: dvel     !!.. perturbed velocity model
  real     , dimension(:,:,:)    , allocatable, private :: deta     !!.. perturbed eta      model
  real     , dimension(:)        , allocatable, private :: bvrf     !!.. reference velocity
  real     , dimension(:,:,:,:,:), allocatable, private :: rimg     !!.. output image, real
  real     , dimension(:,:,:,:,:), allocatable, private :: hess
  real     , dimension(:,:)      , allocatable, private :: imsk
  real     , dimension(:,:,:)    , allocatable, private :: coef     !!.. optimized coefficients
  real     , dimension(:,:,:)    , allocatable, private :: dr_coef  !!.. derivative of optimized coefficients

  complex  , dimension(:)        , allocatable, private :: sfnc
  complex  , dimension(:,:,:,:,:), allocatable, private :: cimg     !!.. output image, complex
  complex  , dimension(:,:)      , allocatable, private :: crec
  complex  , dimension(:,:)      , allocatable, private :: csou
  complex  , dimension(:,:,:,:)  , allocatable, private :: bwfs, bwfr
  complex  , dimension(:,:,:,:)  , allocatable, private :: dwfs, dwfr

  complex  , dimension(:,:,:)    , allocatable, private :: bwfs_slice, bwfr_slice
  complex  , dimension(:,:,:)    , allocatable, private :: dwfs_slice, dwfr_slice
  complex  , dimension(:,:,:)    , allocatable, private :: wtmp_slice

  real     , dimension(:,:) 	 , allocatable, private :: alpha_z, beta_z   !!.. coefficients for depth z 
  real     , dimension(:,:) 	 , allocatable, private :: dalpha_z, dbeta_z !!.. deriv of coefficients for depth z 

  real     , dimension(:,:,:,:)  , allocatable, private :: imag

  real     , dimension(:)        , allocatable, private :: kx2
  real     , dimension(:)        , allocatable, private :: ky2
  real     , dimension(:)        , allocatable, private :: tapx
  real     , dimension(:)        , allocatable, private :: tapy

  integer*8, dimension(:)        , allocatable, private :: plan_2dsF, plan_2dsI
  integer*8, dimension(:)        , allocatable, private :: plan_2drF, plan_2drI
  integer*8, dimension(:)        , allocatable, private :: plan_2dtF, plan_2dtI  

  integer  , dimension(:,:)      , allocatable, private :: index_mxs, index_mxr
  integer  , dimension(:,:)      , allocatable, private :: index_mys, index_myr

  integer  , dimension(:,:,:)    , allocatable, private :: iwrk2d
  real     , dimension(:,:)      , allocatable, private :: rwrk1d
  complex  , dimension(:,:)      , allocatable, private :: cwrk1d
  complex  , dimension(:,:,:,:)  , allocatable, private :: cwrk3d

  integer                                     , private :: nzrept
  integer                                     , private :: bscnop
  character(len=6)                            , private :: bsccoe
  character(len=6)                            , private :: bsctyp
 
  logical                                     , private :: wntwfs
  logical                                     , private :: wntwfr
  logical                                     , private :: report
  logical                                     , private :: memchk

  integer                                     , private :: imgfwd_sign
  logical                                     , private :: imgfwd_oper

  integer                                     , private :: tombwd_sign
  logical                                     , private :: tombwd_oper

  real                                        , private :: eta0  , del0 
  real                                        , private :: alpha0, beta0

  integer                                     , private :: ompchunk, nth
  integer  ,                          external, private :: omp_get_num_procs
  integer  ,                          external, private :: omp_get_num_threads
  integer  ,                          external, private :: omp_set_num_threads
  integer  ,                          external, private :: omp_get_thread_num

  real     ,                         parameter, private :: PI = 3.14159265359

contains


  integer function bwi_kernel_init(dgrid_out, pgrid_out, vgrid_out, igrid_out, hgrid_out, cgrid_out, ftag_out_) result(stat)
    type(dat_grid)   :: dgrid_out
    type(pro_grid)   :: pgrid_out
    type(vel_grid)   :: vgrid_out
    type(img_grid)   :: igrid_out
    type(hes_grid)   :: hgrid_out
    type(cef_grid)   :: cgrid_out
    type(bwi_ftag)   :: ftag_out_

    real             :: bndpr, bndpi
    character(len=7) :: imgfwd_type
    character(len=7) :: tombwd_type

integer :: i

    ftag_%csou_ = "csou"       !!.. source function
    ftag_%crec_ = "crec"       !!.. receiver data
    ftag_%bvel_ = "bvel"       !!.. background velocity
    ftag_%beta_ = "beta"       !!.. background eta
    ftag_%bdel_ = "bdel"       !!.. background delta
    ftag_%dvel_ = "dvel"       !!.. velocity perturbation 
    ftag_%deta_ = "deta"       !!.. eta      perturbation 
    ftag_%bimg_ = "bimg"       !!.. bakground image
    ftag_%dimg_ = "dimg"       !!.. image perturbation
    ftag_%grdo_ = "grdo"       !!.. gradient at current iteration
    ftag_%grdp_ = "grdp"       !!.. gradient at next iteration 
    ftag_%srch_ = "srch"       !!.. search direction 
    ftag_%bwfs_ = "bwfs"       !!.. source wavefield
    ftag_%bwfr_ = "bwfr"       !!.. receiver wavefield
    ftag_%sfnc_ = "sfnc"       !!.. source signature
    ftag_%gsou_ = "gsou"       !!.. source function for source-side Green's function
    ftag_%grec_ = "grec"       !!.. source function for receiver-side Green's function
    ftag_%hess_ = "hess"       !!.. Hessian
    ftag_%imsk_ = "imsk"       !!.. image mask
    ftag_%gmsk_ = "gmsk"       !!.. gradient mask
    ftag_%hsou_ = "hsou"       !!.. source header
	ftag_%coef_ = "coef"	   !!.. optimized anisotropic coefficients for extrapolation	
	ftag_%dcef_ = "dcef"	   !!.. derivative of optimized anisotropic coefficients 	

!   read parameters
!   ---------------
    call from_param("radian", dgrid%radian, .true.  ) !!.. whether the input frequency is in radian or not
    call from_param("offset", dgrid%offset, .false. ) !!.. whether input data are in receiver or offset coordinates
    call from_param("comsou", dgrid%comsou, .false. ) !!.. point source or composite source
    call from_param("hdrsou", dgrid%hdrsou, .false. ) !!.. input source header or not
    call from_param("smtsou", dgrid%smtsou, .false. ) !!.. smooth point source or not
    call from_param("cpximg", igrid%cpximg, .false. ) !!.. output complex image
    call from_param("stkimg", igrid%stkimg, .true.  ) !!.. output image stacked over shot axis
    call from_param("imgmsk", igrid%imgmsk, .false. ) !!.. whether or not to apply image mask
    call from_param("grdmsk", vgrid%grdmsk, .false. ) !!.. whether or not to apply gradient maskt
    call from_param("srcbnd", pgrid%srcbnd, .false. ) !!.. boundary condition for source function, change dipole source into monopole
    call from_param("recbnd", pgrid%recbnd, .false. ) !!.. boundary condition for the receivers. use this only for impluse response test 
    call from_param("vsfc"  , pgrid%vsfc  , 1500.   ) !!.. surface velocity for boundary conditions
    call from_param("ntpx"  , pgrid%ntpx  , 1       ) !!.. taper width in x
    call from_param("ntpy"  , pgrid%ntpy  , 1       ) !!.. taper width in y
    call from_param("nvrf"  , pgrid%nvrf  , 1       ) !!.. number of reference velocities
    call from_param("oper"  , pgrid%oper  , "ssfpi" ) !!.. extrapolator type, phsft/ssfou/impfd/foufd/phspi/[ssfpi]
    call from_param("mode"  , pgrid%mode  , "imgadj") !!.. mode to run
    call from_param("wnthes", hgrid%wnthes, .false. ) !!.. whether or not to output Hessian
    call from_param("difsou", hgrid%difsou, .false. ) !!.. whether or not to use the same source function for both migration and Hessian
    call from_param("grecin", hgrid%grecin, .false. ) !!.. whether or not to input source functions for receiver-side Green's function
    call from_param("hestyp", hgrid%hestyp, 1       ) !!.. Hessian type, 1 for source wavefield intensity; 2 for phase-encoded Hessian
    call from_param("rdmtyp", hgrid%rdmtyp, 2       ) !!.. random phase encoding type, 1 for no encoding; 2 for phase code; 3 for binary code; 4 for random time delay
    call from_param("nws"   , wblock%ns   , 1       ) !!.. number of frequencies to hold in memory
    call from_param("bscnop", bscnop      , 2       ) !!.. number of terms for Born scattering operator
    call from_param("bsctyp", bsctyp      , "fourie") !!.. type of Born scattering, "fourie" for Fourier domain; "tayexp" for Taylor expansion 
    call from_param("bsccoe", bsccoe      , "optimz") !!.. coefficients for Born scattering operator, "taylor" for Taylor coeffs; "optimz" for optimized coeffs
    call from_param("nzrept", nzrept      , 50      ) !!.. report every nzrept depths
    call from_param("wntwfs", wntwfs      , .false. ) !!.. whether or not to output source wavefield
    call from_param("wntwfr", wntwfr      , .false. ) !!.. whether or not to output receiver wavefield
    call from_param("report", report      , .true.  ) !!.. print report or not 
    call from_param("memchk", memchk      , .true.  ) !!.. memory check or not

    call from_param("imgfwd_type", imgfwd_type, "bckward") !!.. extrapolation type for imaging/modeling
    call from_param("tombwd_type", tombwd_type, "adjoint") !!.. extrapolation type for tomography

    select case (imgfwd_type)
      case ("bckward")
           imgfwd_oper = .false.; imgfwd_sign = +1
      case ("adjoint")
           imgfwd_oper = .true. ; imgfwd_sign = -1
      case default
           write(0,*)"imgfwd_type has to be either <bckward> or <adjoint>"
           STOP
    end select

    select case (tombwd_type)
      case ("bckward")
           tombwd_oper = .false.; tombwd_sign = +1
      case ("adjoint")
           tombwd_oper = .true. ; tombwd_sign = -1
      case default
           write(0,*)"tombwd_type has to be either <bckward> or <adjoint>"
           STOP
    end select
    
    if (pgrid%oper=="phspi" .or. pgrid%oper=="ssfpi") then
       pgrid%pdip  = 0
       pgrid%trick = 0.
       pgrid%bndp  = 0.

    else 
       call from_param("pdip"  , pgrid%pdip  , 45     ) !!.. propagation dip for implicit finite difference
       call from_param("trick" , pgrid%trick , 0.14   ) !!.. 1/6 trick
       call from_param("bndpr" , bndpr       , 0.     ) !!.. boundary condition, real part
       call from_param("bndpi" , bndpi       , 0.     ) !!.. boundary condition, imaginary part
       call from_param("bullet", pgrid%bullet, .false.) !!.. use bulletproof scheme or not
       call from_param("errflt", pgrid%errflt, .false.) !!.. apply Li's compensation filter or not
       pgrid%bndp = cmplx(bndpr, bndpi)
    end if

    if (report) then
       write(0,*)"imgfwd_type = ", imgfwd_type
       write(0,*)"imgfwd_oper = ", imgfwd_oper, imgfwd_sign
       write(0,*)"tombwd_type = ", tombwd_type
       write(0,*)"tombwd_oper = ", tombwd_oper, tombwd_sign
       write(0,*)"oper = ", pgrid%oper, " nvrf = ", pgrid%nvrf
    end if


    select case (pgrid%mode)
      case ("imgfwd")
           if (report) then 
              write(0,*) "Imaging: FORWARD"
           end if 
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%beta_)
           call auxinout(ftag_%bdel_)
           call auxinout(ftag_%bimg_)
           call auxinout(ftag_%coef_)
           call bwi_histos_imgfwd(dgrid, pgrid, vgrid, igrid, cgrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_, ftag_%coef_)

      case ("imgadj")
           if (report) then
              write(0,*) "Imaging: ADJOINT"
           end if 
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%beta_)
           call auxinout(ftag_%bdel_)
           call auxinout(ftag_%bimg_)
           call auxinout(ftag_%coef_)
           call bwi_histos_imgadj(dgrid, pgrid, vgrid, igrid, cgrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_, ftag_%hsou_, ftag_%coef_)

      case ("imginv")
      case ("tomfwd")
           if (report) then
              write(0,*) "Tomography: FORWARD"
           end if
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%beta_)
           call auxinout(ftag_%bdel_)
           call auxinout(ftag_%dvel_)
           call auxinout(ftag_%deta_)
           call auxinout(ftag_%dimg_)
           call auxinout(ftag_%coef_)
           call auxinout(ftag_%dcef_)
           call bwi_histos_tomfwd(dgrid, pgrid, vgrid, igrid, cgrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%dvel_, &
								   ftag_%deta_, ftag_%dimg_, ftag_%coef_)

      case ("tomadj") 
           if (report) then
               write(0,*) "Tomography: ADJOINT" 
           end if
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%beta_)
           call auxinout(ftag_%bdel_)
           call auxinout(ftag_%dvel_)
           call auxinout(ftag_%deta_)
           call auxinout(ftag_%dimg_)
           call auxinout(ftag_%coef_)
           call auxinout(ftag_%dcef_)
           call bwi_histos_tomadj(dgrid, pgrid, vgrid, igrid, cgrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%dvel_, &
								  ftag_%deta_, ftag_%dimg_, ftag_%coef_)

      case ("tominv")
           if (report) then
              write(0,*) "Tomography: INVERSION"
           end if
           call auxinout(ftag_%csou_)
           call auxinout(ftag_%crec_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%beta_)
           call auxinout(ftag_%bdel_)
           call auxinout(ftag_%bimg_)
           call auxinout(ftag_%grdo_)
           call auxinout(ftag_%grdp_)
           call auxinout(ftag_%srch_)
           call auxinout(ftag_%coef_)
           call auxinout(ftag_%dcef_)
           if (hgrid%wnthes) then
              call auxinout(ftag_%hess_)
              if (hgrid%difsou) then
                 call auxinout(ftag_%gsou_)
              else
                 ftag_%gsou_ = ftag_%csou_
              end if
              if (hgrid%grecin   ) call auxinout(ftag_%grec_)
              if (hgrid%hestyp==2) call auxinout(ftag_%sfnc_)
           end if
           if (igrid%imgmsk) call auxinout(ftag_%imsk_)
           if (vgrid%grdmsk) call auxinout(ftag_%gmsk_)
           call bwi_histos_tominv(dgrid, pgrid, vgrid, igrid, cgrid, hgrid, ftag_%csou_, ftag_%crec_, ftag_%bvel_, &
                                  ftag_%bimg_, ftag_%hess_, ftag_%grdo_, ftag_%grdp_, ftag_%srch_, ftag_%coef_) 

      case ("hesian")
           if (report) then
               write(0,*) "HESSIAN"
           end if 
           call auxinout(ftag_%gsou_)
           call auxinout(ftag_%hess_)
           call auxinout(ftag_%bvel_)
           call auxinout(ftag_%beta_)
           call auxinout(ftag_%bdel_)
           call auxinout(ftag_%coef_)
           if (hgrid%grecin   ) call auxinout(ftag_%grec_)
           if (hgrid%hestyp==2) call auxinout(ftag_%sfnc_)
           call bwi_histos_hesian(dgrid, pgrid, vgrid, igrid, cgrid, hgrid, ftag_%sfnc_, ftag_%gsou_, ftag_%grec_, ftag_%bvel_, ftag_%hess_, ftag_%coef_)

    end select

!   setup OpenMP stuff
!   ------------------
    ompchunk = 1
    nth      = 1

    !$OMP PARALLEL
    if (omp_get_thread_num() == 0) then
       nth = omp_get_num_threads()
       if (report) then 
          write(0,*)"nthread =", nth, omp_get_num_procs()
       end if
    end if
    !$OMP END PARALLEL

!   radian
!   ------
    if (.not.dgrid%radian) then
       dgrid%aw%d = dgrid%aw%d*2*PI
       dgrid%aw%o = dgrid%aw%o*2*PI
    end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!   if present the source header, read it
!   -------------------------------------
if (dgrid%hdrsou) then
   allocate(dgrid%hsou(2,dgrid%nshg))
   call sep_read(dgrid%hsou, ftag_%hsou_)
end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


!   calculate the global parameters for the first shot
!   --------------------------------------------------
    call bwi_glocal(pgrid, ogrid, vgrid, dgrid, igrid, hgrid, 1)

!   calculate the frequency blocks
!   ------------------------------
    call base_blkcal(wblock, dgrid%aw)

!   check memory
!   ------------
    call bwi_memory_chck(memchk)

!   allocate arrays
!   ---------------
    call bwi_arrays_init()

!   initialize modules
!   ------------------
    call bwi_module_init()

    if (wntwfs) then
       call auxinout(ftag_%bwfs_)
       call bwi_histos_wfld(dgrid, pgrid, ftag_%bwfs_)
    end if
    if (wntwfr) then
       call auxinout(ftag_%bwfr_)
       call bwi_histos_wfld(dgrid, pgrid, ftag_%bwfr_)
    end if

    dgrid_out = dgrid
    pgrid_out = pgrid
    vgrid_out = vgrid   !! bug danger
    igrid_out = igrid
    hgrid_out = hgrid
	cgrid_out = cgrid
    ftag_out_ = ftag_

    stat = 0
  end function

  integer function bwi_kernel_clos() result(stat)
    stat = 0
  end function


  integer function bwi_kertom_oper(adj, add, csou_, crec_, bvel_, beta_, bdel_, dvel_, deta_, dimg_) result(stat)
    logical           :: adj, add
    character(len=*)  :: csou_, crec_, bvel_, beta_, bdel_, dimg_, dvel_, deta_

    type(axis)        :: aws
    integer           :: izs, ish, ishx, ishy
    integer           :: ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak

    if (.not.add) then
       if (adj) then
          call bwi_fio_dvel_init(vgrid, dvel_)
          call bwi_fio_dvel_init(vgrid, deta_)
       else
          call bwi_fio_rimg_init(igrid, dimg_)
       end if
    end if

!	read the coefficients and the derivative of them
!	----------------------
	call bwi_fio_coef(coef   ,cgrid,ftag_%coef_)	
	call bwi_fio_coef(dr_coef,cgrid,ftag_%dcef_)	

!   loop over shots
!   ---------------
    do ish=1, dgrid%nshg
!      output image coordinates
!      ------------------------
       if (igrid%stkimg) then
          ishx = 1
          ishy = 1
       else
          ishx = mod((ish-1)/1          , dgrid%asx%n)+1
          ishy = mod((ish-1)/dgrid%asx%n, dgrid%asy%n)+1
       end if
!      some backups
!      ------------
       ogrid_nx_bak = ogrid%ax%n; ogrid_ny_bak = ogrid%ay%n; ogrid_nz_bak = ogrid%az%n
!      calculate the global parameters
!      -------------------------------
       call bwi_glocal(pgrid, ogrid, vgrid, dgrid, igrid, hgrid, ish)
       if (igrid%ax%n>1.and.ogrid%ax%n == 0) cycle
       if (igrid%ay%n>1.and.ogrid%ay%n == 0) cycle
       call bwi_arrays_rset(ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak)
       if (report) call bwi_report(dgrid, vgrid, pgrid, ogrid, igrid, ish)
!      initialize extrapolators
!      ------------------------
       stat = ani_wfdexp_init(kx2,ky2,pgrid%ax%n,pgrid%ay%n,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%pdip,&
                              pgrid%nvrf,pgrid%oper,pgrid%trick,pgrid%bndp,pgrid%bullet,pgrid%errflt)
!      initialize reference velocity
!      -----------------------------
       stat = bwi_velref_init(pgrid%nvrf)
!      initialize imaging condition
!      ----------------------------
       stat = bwi_imgcon_init(ogrid, pgrid, dgrid%aw%n, dgrid%nshg, index_mxs, index_mxr, index_mys, index_myr)
!      initialize get coefficients  
!      ----------------------------
	   stat = ani_getcef_init(cgrid, pgrid, coef, dr_coef)	
!      initialize the Born scattering operator
!      ---------------------------------------
       stat = ani_bornsc_init(kx2,ky2,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%ax%n,pgrid%ay%n,pgrid%nvrf,bscnop,bsctyp,bsccoe)

!stat = bwi_bornsc_init(kx2,ky2,pgrid%ax%d,pgrid%ay%d,pgrid%az%d,pgrid%ax%n,pgrid%ay%n,bscnop,2)
!      apply the tomographic operator
!      ------------------------------
       call bwi_tom1st(adj, add, csou_, crec_, bvel_, beta_, bdel_, dvel_, deta_, dimg_, ish, ishx, ishy)
!      close subroutines
!      -----------------
       stat = ani_bornsc_clos()
       stat = bwi_imgcon_clos()
       stat = bwi_velref_clos()
       stat = ani_wfdexp_clos() 
	   stat = ani_getcef_clos()
    end do
    stat = 0
  end function

  subroutine bwi_tom1st(adj, add, csou_, crec_, bvel_, beta_, bdel_, dvel_, deta_, dimg_, ish, ishx, ishy)
    logical          :: adj, add
    character(len=*) :: csou_, crec_, bvel_, beta_, bdel_, dvel_, deta_, dimg_
    integer          :: ish, ishx, ishy 

    type(axis)       :: aws
    integer          :: iwb, iws, izs, jzs, ith, stat
    real             :: ws


    if (adj) then     !!.. dS = F' dI
       do iwb=1, wblock%nb
          if (report) write(0,*)"Working at iwb = ", iwb, "out of ", wblock%nb
          aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)
          bwfs = 0.; bwfr = 0.
          dwfs = 0.; dwfr = 0.
          call bwi_fio_csou(csou, dgrid, wblock%indx(iwb), aws%n, ish, csou_, "reed")
          call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, crec_, "reed")
          call bwi_wfs_init(.true., .false., bwfs(:,:,:,1), csou, dgrid, pgrid)
          call bwi_wfr_init(.true., .false., bwfr(:,:,:,1), crec, dgrid, pgrid)

!         apply surface boundary condition for source (change dipole source into mono-pole)
!         ---------------------------------------------------------------------------------
          if (pgrid%srcbnd) then
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws) 
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,1)
                call bwi_sbcond(dwfs_slice(:,:,ith), kx2, ky2, pgrid%vsfc, ws, pgrid%ax%n, pgrid%ay%n, plan_2dsF(ith), plan_2dsI(ith))
                bwfs(:,:,iws,1) = dwfs_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end if

!         compute the background wavefields
!         --------------------------------- 
          do izs=2, pgrid%nzstep
             call bwi_fio_velo(bvel, vgrid, pgrid, izs-1, bvel_, "reed")
             call bwi_fio_velo(beta, vgrid, pgrid, izs-1, beta_, "reed")
             call bwi_fio_velo(bdel, vgrid, pgrid, izs-1, bdel_, "reed")
			 
!            get the reference velocity eta and delta 
!            -------------------
             stat = bwi_velref_oper(bvel, bvrf)
			 eta0 = sum(beta) / size(beta)
			 del0 = sum(bdel) / size(bdel)

!            get the coefficients at depth z for implicit extrapolation 
!            -------------------
			 stat = ani_getcef_oper (beta, bdel, alpha_z , beta_z)	
			 stat = ani_getcef_oper2(eta0, del0, alpha0  , beta0 )	

             dwfs_slice = 0.; dwfr_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs-1)
                dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs-1)
				stat = ani_wfdexp_oper(.false.,-1,dwfs_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith))
                stat = ani_wfdexp_oper(.false.,-1,dwfr_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith))
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                bwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)
                bwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end do

          !!.. dP_i = E_{i}'(dP_{i-1}+G_{i-1}dV_{i-1})
          do izs=pgrid%nzstep-1, 1, -1
!            read background velocity at current depth step: bV_{i} 
!            ------------------------------------------------------
             call bwi_fio_velo(bvel, vgrid, pgrid, izs, bvel_, "reed")
             call bwi_fio_velo(beta, vgrid, pgrid, izs, beta_, "reed")
             call bwi_fio_velo(bdel, vgrid, pgrid, izs, bdel_, "reed")
!            calculate the reference velocity, eta and delta
!            -------------------------------- 
             stat = bwi_velref_oper(bvel, bvrf)
			 eta0 = sum(beta) / size(beta)
			 del0 = sum(bdel) / size(bdel)
!            get the coefficients at depth z for implicit extrapolation 
!            -------------------
	  		 stat = ani_getcef_oper (beta, bdel, alpha_z, beta_z, dalpha_z, dbeta_z)	
			 stat = ani_getcef_oper2(eta0, del0, alpha0,  beta0 )	
!            initialize arrays
!            -----------------
             rimg = 0.; dvel = 0.; deta = 0.
!            read the perturbed image at previous depth level: dI_{i+1}
!            ----------------------------------------------------------
             if (izs+1>=ogrid%index_outpro_zmin .and. izs+1<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                call bwi_fio_rimg(rimg(:,:,:,:,1), igrid, ogrid, jzs+1, ishx, ishy, dimg_, "reed")
             end if
!            copy it to all threads
!            ----------------------
             do ith=1, nth
                rimg(:,:,:,:,ith) = rimg(:,:,:,:,1)
             end do
!            read the existing perturbed velocity at current depth level for accumulating dV_{i}
!            -----------------------------------------------------------------------------------
             call bwi_fio_velo(dvel(:,:,1), vgrid, pgrid, izs, dvel_, "reed")
!            read the existing perturbed eta at current depth level for accumulating dV_{i}
!            -----------------------------------------------------------------------------------
             call bwi_fio_velo(deta(:,:,1), vgrid, pgrid, izs, deta_, "reed")
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
!               initialize buffers
!               ------------------
                bwfs_slice(:,:,ith) = 0.; bwfr_slice(:,:,ith) = 0.
                dwfs_slice(:,:,ith) = 0.; dwfr_slice(:,:,ith) = 0.
                wtmp_slice(:,:,ith) = 0.

!               ************************ 
!               SOURCE-PART PERTURBATION
!               ************************
!               perturbed auxiliary wavefield at previous depth level: dP_{i+1}
!               ----------------------------------------------------------------------
                dwfs_slice(:,:,ith) = dwfs(:,:,iws,izs+1)
!               accumulate the perturbed auxiliary wavefield with the perturbed source wavefield: dP_{i+1} = dP_{i+1} + bU_{i+1}dI_{i+1}
!               ------------------------------------------------------------------------------------------------------------------------
                if (izs+1>=ogrid%index_outpro_zmin .and. izs+1<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.true., rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), bwfr(:,:,iws,izs+1), ws, "sou")
                end if
!               taper the perturbed auxiliary wavefield
!               ---------------------------------------
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               propagate the perturbed auxiliary wavefield to current depth level: dP_{i} = E_{i}' dP_{i+1}
!               --------------------------------------------------------------------------------------------
				stat = ani_wfdexp_oper(.false.,+1,dwfs_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith))
!               compute the scattering operator at previous depth step: G_{i-1} for velocity
!               ----------------------------------------------------------------------------
				bwfs_slice(:,:,ith) = 0.
                stat = ani_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs),wtmp_slice(:,:,ith),&
                                       bvel,beta,bdel,bvrf,alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"vel")
!               compute the velocity perturbation at current depth level: dV_{i} =  G_{i}' dP_{i}
!               ---------------------------------------------------------------------------------
                dvel(:,:,ith) = dvel(:,:,ith) + conjg(bwfs_slice(:,:,ith))*dwfs_slice(:,:,ith)

!               compute the scattering operator at previous depth step: G_{i} for eta
!               ----------------------------------------------------------------------------
				bwfs_slice(:,:,ith) = 0.
                stat = ani_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs),wtmp_slice(:,:,ith),&
                                       bvel,beta,bdel,bvrf,alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"eta",dalpha_z,dbeta_z)
!               compute the velocity perturbation at current depth level: dV_{i} =  G_{i}' dP_{i}
!               ---------------------------------------------------------------------------------
                deta(:,:,ith) = deta(:,:,ith) + conjg(bwfs_slice(:,:,ith))*dwfs_slice(:,:,ith)
!               update the perturbed auxiliary wavefield: dP_{i}
!               ------------------------------------------------
                dwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)

!               **************************
!               RECEIVER_PART PERTURBATION
!               **************************
!               perturbed auxiliary wavefield at previous depth level: dP_{i+1}
!               ---------------------------------------------------------------
                dwfr_slice(:,:,ith) = dwfr(:,:,iws,izs+1)
!               accumulate the perturbed auxiliary wavefield with the perturbed receiver wavefield: dP_{i+1} = dP_{i+1} + bD_{i+1}dI_{i+1}
!               --------------------------------------------------------------------------------------------------------------------------
                if (izs+1>=ogrid%index_outpro_zmin .and. izs+1<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.true., rimg(:,:,:,:,ith), bwfs(:,:,iws,izs+1), dwfr_slice(:,:,ith), ws, "rec")
                end if
!               taper the perturbed receiver wavefield
!               --------------------------------------
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               propagate the perturbed auxiliary wavefield to current depth level: dP_{i} = E{i}' dP_{i+1}
!               -------------------------------------------------------------------------------------------
				stat = ani_wfdexp_oper(.false.,+1,dwfr_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith))
!               compute the scattering operator at previous depth step: G_{i} for velocity
!               ----------------------------------------------------------------------------
				bwfr_slice(:,:,ith) = 0.
                stat = ani_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs),wtmp_slice(:,:,ith),&
                                       bvel,beta,bdel,bvrf,alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"vel")
!               compute the velocity perturbation at current depth level: dV_{i} = G_{i}' dP_{i}
!               --------------------------------------------------------------------------------
                dvel(:,:,ith) = dvel(:,:,ith) + conjg(bwfr_slice(:,:,ith))*dwfr_slice(:,:,ith)

!               compute the scattering operator at previous depth step: G_{i} for eta
!               ----------------------------------------------------------------------------
				bwfr_slice(:,:,ith) = 0.
                stat = ani_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs),wtmp_slice(:,:,ith),&
                                       bvel,beta,bdel,bvrf,alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"eta",dalpha_z,dbeta_z)
!               compute the velocity perturbation at current depth level: dV_{i} =  G_{i}' dP_{i}
!               ---------------------------------------------------------------------------------
                deta(:,:,ith) = deta(:,:,ith) + conjg(bwfr_slice(:,:,ith))*dwfr_slice(:,:,ith)
!               update the perturbed auxiliary wavefield: dP_{i}
!               ------------------------------------------------
                dwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)


             end do
             !$OMP END PARALLEL DO

!            write out veloicty perturbation
!            -------------------------------
		     dvel(:,:,1) = sum(dvel,3); deta(:,:,1) = sum(deta,3);
             call bwi_fio_velo(dvel(:,:,1), vgrid, pgrid, izs, dvel_, "rite")
             call bwi_fio_velo(deta(:,:,1), vgrid, pgrid, izs, deta_, "rite")
          end do
       end do

    else              !!.. dI = F  dS

       do iwb=1, wblock%nb
          if (report) write(0,*)"Working at iwb = ", iwb, "out of ", wblock%nb
          aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)
          bwfs = 0.; bwfr = 0.; 
		  dwfs = 0.; dwfr = 0.
          call bwi_fio_csou(csou, dgrid, wblock%indx(iwb), aws%n, ish, csou_, "reed")
          call bwi_fio_crec(crec, dgrid, wblock%indx(iwb), aws%n, ish, crec_, "reed")
          call bwi_wfs_init(.true., .false., bwfs(:,:,:,1), csou, dgrid, pgrid)
          call bwi_wfr_init(.true., .false., bwfr(:,:,:,1), crec, dgrid, pgrid)

!         apply surface boundary condition for source (change dipole source into mono-pole)
!         ---------------------------------------------------------------------------------
          if (pgrid%srcbnd) then
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws) 
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,1)
                call bwi_sbcond(dwfs_slice(:,:,ith), kx2, ky2, pgrid%vsfc, ws, pgrid%ax%n, pgrid%ay%n, plan_2dsF(ith), plan_2dsI(ith))
                bwfs(:,:,iws,1) = dwfs_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end if

!         compute the backougrnd source and receiver wavefields
!         -----------------------------------------------------
          do izs=2, pgrid%nzstep
             call bwi_fio_velo(bvel, vgrid, pgrid, izs-1, bvel_, "reed")
             call bwi_fio_velo(beta, vgrid, pgrid, izs-1, beta_, "reed")
             call bwi_fio_velo(bdel, vgrid, pgrid, izs-1, bdel_, "reed")
             stat = bwi_velref_oper(bvel, bvrf)
			 eta0 = sum(beta) / size(beta)
			 del0 = sum(bdel) / size(bdel)

!            get the coefficients at depth z for implicit extrapolation 
!            -------------------
			 stat = ani_getcef_oper (beta, bdel, alpha_z, beta_z)	
			 stat = ani_getcef_oper2(eta0, del0, alpha0,  beta0 )	

             dwfs_slice = 0.; dwfr_slice = 0.
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
                dwfs_slice(:,:,ith) = bwfs(:,:,iws,izs-1)
                dwfr_slice(:,:,ith) = bwfr(:,:,iws,izs-1)
				stat = ani_wfdexp_oper(.false.,-1,dwfs_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith))
                stat = ani_wfdexp_oper(.false.,-1,dwfr_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2drF(ith),plan_2drI(ith))
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
                bwfs(:,:,iws,izs) = dwfs_slice(:,:,ith)
                bwfr(:,:,iws,izs) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO
          end do

          !!.. dP_i = E_{i-1}(dP_{i-1}+G_{i-1}dS_{i-1})
          do izs=1, pgrid%nzstep-1
!            read the background velocity at previous depth level for extrapolation bV_{i-1} => E_{i-1}
!            ------------------------------------------------------------------------------------------
             call bwi_fio_velo(bvel, vgrid, pgrid, izs, bvel_, "reed")
             call bwi_fio_velo(beta, vgrid, pgrid, izs, beta_, "reed")
             call bwi_fio_velo(bdel, vgrid, pgrid, izs, bdel_, "reed")
!            calculate the reference velocity, eta and delta
!            -------------------------------- 
             stat = bwi_velref_oper(bvel, bvrf)
			 eta0 = sum(beta) / size(beta)
			 del0 = sum(bdel) / size(bdel)
!            get the coefficients at depth z for implicit extrapolation 
!            -------------------
	  		 stat = ani_getcef_oper (beta, bdel, alpha_z, beta_z, dalpha_z, dbeta_z)	
			 stat = ani_getcef_oper2(eta0, del0, alpha0,  beta0 )	
!            initialize some buffers
!            -----------------------
             rimg = 0.; dvel = 0.; deta = 0.
!            read pertubed image at current depth level dI_{i}
!            -------------------------------------------------
             if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                call bwi_fio_rimg(rimg(:,:,:,:,1), igrid, ogrid, jzs+1, ishx, ishy, dimg_, "reed")
             end if
!            read the velocity perturbation at previous depth level dV_{i-1}
!            ---------------------------------------------------------------
             call bwi_fio_velo(dvel(:,:,1), vgrid, pgrid, izs, dvel_, "reed")      !!.. dS_{i-1}
!            read the eta perturbation at previous depth level dE_{i-1}
!            ---------------------------------------------------------------
             call bwi_fio_velo(deta(:,:,1), vgrid, pgrid, izs, deta_, "reed")      !!.. dE_{i-1}
!            copy it to all threads
!            ----------------------
             do ith=1, nth
                dvel(:,:,ith) = dvel(:,:,1)
                deta(:,:,ith) = deta(:,:,1)
             end do
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,izs,ith,ws)
             do iws=1, aws%n
                ith = omp_get_thread_num() + 1
                ws  = aws%o + (iws-1)*aws%d
!               initilize buffers
!               -----------------
                bwfs_slice(:,:,ith) = 0.; bwfr_slice(:,:,ith) = 0.
                dwfs_slice(:,:,ith) = 0.; dwfr_slice(:,:,ith) = 0.
                wtmp_slice(:,:,ith) = 0.

!               ************************
!               SOURCE-PART PERTURBATION 
!               ************************
!               perturbed source wavefield at previous depth level: dD_{i-1}
!               ------------------------------------------------------------
                dwfs_slice(:,:,ith) = dwfs(:,:,iws,izs)
!               compute the scattering operator at previous depth step: G_{i-1} for velocity
!               ----------------------------------------------------------------------------
                stat = ani_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs),wtmp_slice(:,:,ith),&
                                       bvel,beta,bdel,bvrf,alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"vel")
!               accumulate the perturbed source wavefield at previous depth level: dD_{i-1} + G_{i-1}dV_{i-1}
!               ---------------------------------------------------------------------------------------------
                dwfs_slice(:,:,ith) = dwfs_slice(:,:,ith) + bwfs_slice(:,:,ith)*dvel(:,:,ith) 
!               compute the scattering operator at previous depth step: G_{i-1} for velocity
!               ----------------------------------------------------------------------------
                stat = ani_bornsc_oper(bwfs_slice(:,:,ith),bwfs(:,:,iws,izs),wtmp_slice(:,:,ith),bvel,beta,bdel,bvrf,&
								alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"eta",dalpha_z,dbeta_z)
!               accumulate the perturbed source wavefield at previous depth level: dD_{i-1} + G_{i-1}dE_{i-1}
!               ---------------------------------------------------------------------------------------------
                dwfs_slice(:,:,ith) = dwfs_slice(:,:,ith) + bwfs_slice(:,:,ith)*deta(:,:,ith) 
!               propagate the accumulated perturbed source wavefield to current depth level: dD_{i} = E_{i-1}[dD_{i-1} + G_{i-1}dV_{i-1}]
!               -------------------------------------------------------------------------------------------------------------------------
				stat = ani_wfdexp_oper(.false.,-1,dwfs_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith))
!               taper the perturbed source wavefield
!               ------------------------------------
                stat = bwi_tapers_oper(dwfs_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               correlate perturbed source wavefield with background receiver wavefield at current depth level: dI_{i} = dD_{i}*bU_{i}
!               ----------------------------------------------------------------------------------------------------------------------
                if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.false., rimg(:,:,:,:,ith), dwfs_slice(:,:,ith), bwfr(:,:,iws,izs+1), ws, "sou")
                end if
!               update the wave matrix dD_{i}
!               -----------------------------
                dwfs(:,:,iws,izs+1) = dwfs_slice(:,:,ith)

!               **************************
!               RECEIVER-PART PERTURBATION
!               **************************
!               perturbed receiver wavefield at previous depth level: dU_{i-1}
!               -------------------------------------------------------------- 
                dwfr_slice(:,:,ith) = dwfr(:,:,iws,izs)
!               compute the scattering operator at previous depth step: G_{i-1} for velocity
!               ---------------------------------------------------------------
                stat = ani_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs),wtmp_slice(:,:,ith),&
                                       bvel,beta,bdel,bvrf,alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"vel")
!               accumulate the perturbed receiver wavefield at previous depth level: dU_{i-1} + G_{i-1}dV_{i-1}
!               -----------------------------------------------------------------------------------------------
                dwfr_slice(:,:,ith) = dwfr_slice(:,:,ith) + bwfr_slice(:,:,ith)*dvel(:,:,ith)
!               compute the scattering operator at previous depth step: G_{i-1} for eta
!               ----------------------------------------------------------------------------
                stat = ani_bornsc_oper(bwfr_slice(:,:,ith),bwfr(:,:,iws,izs),wtmp_slice(:,:,ith),bvel,beta,bdel,bvrf,&
								alpha_z,beta_z,iwrk2d(:,:,ith),ws,-1,plan_2dtF(ith),plan_2dtI(ith),"eta",dalpha_z,dbeta_z)
!               accumulate the perturbed source wavefield at previous depth level: dD_{i-1} + G_{i-1}dV_{i-1}
!               ---------------------------------------------------------------------------------------------
                dwfr_slice(:,:,ith) = dwfr_slice(:,:,ith) + bwfr_slice(:,:,ith)*deta(:,:,ith) 
!               propagate the accumulated perturbed source wavefield to current depth level: dU_{i} = E_{i-1}[dU_{i-1} + G_{i-1}dV_{i-1}]
!               -------------------------------------------------------------------------------------------------------------------------
				stat = ani_wfdexp_oper(.false.,-1,dwfr_slice(:,:,ith),bvel,eta0,del0,alpha_z,beta_z,bvrf,ws,alpha0,beta0,iwrk2d(:,:,ith),rwrk1d(:,ith), &
                                          cwrk1d(:,ith),cwrk3d(:,:,:,ith),plan_2dsF(ith),plan_2dsI(ith))
!               taper the perturbed receiver wavefield
!               --------------------------------------
                stat = bwi_tapers_oper(dwfr_slice(:,:,ith), pgrid%ax%n, pgrid%ay%n)
!               correlate perturbed receiver wavefield with background source wavefield at current depth level: dI_{i} = dU_{i}*bD_{i}
!               ----------------------------------------------------------------------------------------------------------------------
                if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                   stat = bwi_imgcon_oper(.false., rimg(:,:,:,:,ith), bwfs(:,:,iws,izs+1), dwfr_slice(:,:,ith), ws, "rec")
                end if
!               update the wave matrix dU_{i}
!               ----------------------------- 
                dwfr(:,:,iws,izs+1) = dwfr_slice(:,:,ith)
             end do
             !$OMP END PARALLEL DO

!            write out the perturbed image
!            -----------------------------
             if (izs>=ogrid%index_outpro_zmin .and. izs<=ogrid%index_outpro_zmax) then
                jzs = izs-ogrid%index_outpro_zmin+1
                call bwi_fio_rimg(sum(rimg,5), igrid, ogrid, jzs+1, ishx, ishy, dimg_, "rite")
             end if
          end do

       end do

    end if
    return
  end subroutine
  
  subroutine bwi_memory_chck(memchk)
    logical          :: memchk
    double precision :: memory, s

    memory = 0
    s      = 1024. !! each term should be divided by s three times to give GB unit
!   arrays needed for all modes
!   ---------------------------
    if (pgrid%oper=="impfd" .or. pgrid%oper=="foufd") then
       memory = memory + (pgrid%ax%n*5/s+pgrid%ay%n*5/s)*nth/s*8/s + (pgrid%ax%n/s+pgrid%ay%n/s)*nth/s*4./s
    end if
    if (pgrid%oper=="phspi" .or. pgrid%oper=="ssfpi") then
       memory = memory + (pgrid%ax%n/s*pgrid%ay%n/s*2*nth/s)*8 + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*2
    end if
    memory = memory + (pgrid%ax%n/s+pgrid%ay%n/s)*4/s/s + (pgrid%ntpx/s+pgrid%ntpy/s)*4/s/s

    select case (pgrid%mode)

      case ("imgfwd")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + (pgrid%nvrf/s)*4/s/s                   &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*2*8                                    &
                  + (nth/s)*4*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s 

           if (igrid%cpximg) then
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*8*s*s
           else
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s
           end if

      case ("imgadj")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &   
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + (pgrid%nvrf/s)*4/s/s                   &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*2*8                                    &
                  + (nth/s)*4*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s
           if (igrid%cpximg) then
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*8*s*s
           else
              memory = memory + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s
           end if

      case ("tomfwd")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s                     &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*4                                      &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*5*8                                    &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s  &
                  + (nth/s)*6*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s

      case ("tomadj")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s                     &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*4                                      &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*5*8                                    &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s  &
                  + (nth/s)*6*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s

      case ("tominv")
           memory = memory + (dgrid%ntrs/s*wblock%ns/s)*8/s + (dgrid%ntrg/s*wblock%ns/s)*8/s &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s                     &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*4                                      &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*2*8*s             &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*5*8                                    &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2*2/s  &
                  + (nth/s)*6*8/s/s                                                          &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s      &
                  + (igrid%ahx%n/s*igrid%ahy%n/s*igrid%ax%n/s*igrid%ay%n/s)*4*s              &
                  + (vgrid%ax%n/s*vgrid%ay%n/s*vgrid%az%n/s)*3*4                       

           if (hgrid%wnthes) then
              memory = memory + (hgrid%ahx_modl%n/s*hgrid%ahy_modl%n/s*hgrid%ax_modl%n/s*hgrid%ay_modl%n/s)*4*s
           end if

           if (igrid%imgmsk) then
              memory = memory + (igrid%ax%n/s*igrid%ay%n/s)*4/s
           end if
 
      case ("hesian")
           memory = memory + wblock%ns/s*8/s/s + dgrid%ntrs/s*wblock%ns/s*8/s            &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*8*s           &
                  + (pgrid%ax%n/s*pgrid%ay%n/s)*4/s + pgrid%nvrf/s*4/s/s                 &
                  + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*8                                  &
                  + (nth/s)*2*8/s/s                                                      &
                  + (ogrid%ahx%n/s*ogrid%ax%n/s)*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2/s  &
                  + (ogrid%ahx%n/s*ogrid%ahy%n/s*ogrid%ax%n/s*ogrid%ay%n/s*nth/s)*4*s*s 

           if (hgrid%hestyp==2) then
              memory = memory + (dgrid%ntrg/s*wblock%ns/s)*8/s                             &
                     + (pgrid%ax%n/s*pgrid%ay%n/s*wblock%ns/s*pgrid%nzstep/s)*8*s          &
                     + (pgrid%ax%n/s*pgrid%ay%n/s*nth/s)*8                                 &
                     + (nth/s)*2*8/s/s                                                     &
                     + (ogrid%ahx%n/s*ogrid%ax%n/s)*2/s + (ogrid%ahy%n/s*ogrid%ay%n/s)*2/s 
           end if

    end select

    if (memchk) then
       write(0,*) "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
       write(0,*) "                                                           "
       write(0,*) "        Predicted memory usage: ", memory, " GB"
       write(0,*) "                                                           "
       write(0,*) "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++" 
       STOP
    end if

    return
  end subroutine



  subroutine bwi_arrays_init()

!   arrays needed for all modes
!   ---------------------------
    if (pgrid%oper=="impfd" .or. pgrid%oper=="foufd") then
       allocate(cwrk1d(pgrid%ax%n*5+pgrid%ay%n*5,nth)) !!.. ifd or ffd
       allocate(rwrk1d(pgrid%ax%n+pgrid%ay%n,nth))     !!.. ifd or ffd
    end if
    if (pgrid%oper=="phspi" .or. pgrid%oper=="ssfpi") then
       allocate(cwrk3d(pgrid%ax%n,pgrid%ay%n,2,nth))   !!.. phs/ssf-epi
    end if

    allocate(iwrk2d(pgrid%ax%n,pgrid%ay%n,nth)) !!.. phs/ssf-epi and Born PSPI 

    allocate(kx2(pgrid%ax%n),ky2(pgrid%ay%n))
    allocate(tapx(pgrid%ntpx),tapy(pgrid%ntpy))

    select case (pgrid%mode)

      case ("imgfwd")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(beta(pgrid%ax%n,pgrid%ay%n))
           allocate(bdel(pgrid%ax%n,pgrid%ay%n))
		   allocate(coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(alpha_z(pgrid%ax%n,pgrid%ay%n), beta_z(pgrid%ax%n,pgrid%ay%n))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           if (igrid%cpximg) then
              allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           else
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           end if

      case ("imgadj")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(beta(pgrid%ax%n,pgrid%ay%n))
           allocate(bdel(pgrid%ax%n,pgrid%ay%n))
		   allocate(coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(alpha_z(pgrid%ax%n,pgrid%ay%n), beta_z(pgrid%ax%n,pgrid%ay%n))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           if (igrid%cpximg) then
              allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           else
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           end if

      case ("tomfwd")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(beta(pgrid%ax%n,pgrid%ay%n))
           allocate(bdel(pgrid%ax%n,pgrid%ay%n))
		   allocate(coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(dr_coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(alpha_z(pgrid%ax%n,pgrid%ay%n), beta_z(pgrid%ax%n,pgrid%ay%n))
		   allocate(dalpha_z(pgrid%ax%n,pgrid%ay%n), dbeta_z(pgrid%ax%n,pgrid%ay%n))
           allocate(dvel(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(deta(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(dwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(wtmp_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(plan_2dtF(nth),plan_2dtI(nth))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))

      case ("tomadj")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(beta(pgrid%ax%n,pgrid%ay%n))
           allocate(bdel(pgrid%ax%n,pgrid%ay%n))
		   allocate(coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(dr_coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(alpha_z(pgrid%ax%n,pgrid%ay%n), beta_z(pgrid%ax%n,pgrid%ay%n))
		   allocate(dalpha_z(pgrid%ax%n,pgrid%ay%n), dbeta_z(pgrid%ax%n,pgrid%ay%n))
           allocate(dvel(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(deta(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(dwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(wtmp_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(plan_2dtF(nth),plan_2dtI(nth))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))

      case ("tominv")
           allocate(csou(dgrid%ntrs,wblock%ns),crec(dgrid%ntrg,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(beta(pgrid%ax%n,pgrid%ay%n))
           allocate(bdel(pgrid%ax%n,pgrid%ay%n))
		   allocate(coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(dr_coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(alpha_z(pgrid%ax%n,pgrid%ay%n), beta_z(pgrid%ax%n,pgrid%ay%n))
		   allocate(dalpha_z(pgrid%ax%n,pgrid%ay%n), dbeta_z(pgrid%ax%n,pgrid%ay%n))
           allocate(dvel(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(deta(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(dwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(bwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(wtmp_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(plan_2drF(nth),plan_2drI(nth))
           allocate(plan_2dtF(nth),plan_2dtI(nth))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
           allocate(imag(igrid%ahx%n,igrid%ahy%n,igrid%ax%n,igrid%ay%n))

           if (hgrid%wnthes) then
              if (hgrid%hestyp==2) allocate(sfnc(wblock%ns))
           end if

           if (igrid%imgmsk) allocate(imsk(igrid%ax%n,igrid%ay%n))

      case ("hesian")
           allocate(sfnc(wblock%ns))
           allocate(csou(dgrid%ntrs,wblock%ns))
           allocate(bwfs(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
           allocate(bvel(pgrid%ax%n,pgrid%ay%n),bvrf(pgrid%nvrf))
           allocate(beta(pgrid%ax%n,pgrid%ay%n))
           allocate(bdel(pgrid%ax%n,pgrid%ay%n))
		   allocate(coef(cgrid%step%n,cgrid%cfd%n,cgrid%cfe%n))
		   allocate(alpha_z(pgrid%ax%n,pgrid%ay%n), beta_z(pgrid%ax%n,pgrid%ay%n))
           allocate(dwfs_slice(pgrid%ax%n,pgrid%ay%n,nth))
           allocate(plan_2dsF(nth),plan_2dsI(nth))
           allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
           allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))

           if (hgrid%hestyp==2) then
              allocate(crec(dgrid%ntrg,wblock%ns))
              allocate(bwfr(pgrid%ax%n,pgrid%ay%n,wblock%ns,pgrid%nzstep))
              allocate(dwfr_slice(pgrid%ax%n,pgrid%ay%n,nth))
              allocate(plan_2drF(nth),plan_2drI(nth))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n)) 
           end if

    end select
    return
  end subroutine

  subroutine bwi_arrays_rset(ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak)
    integer :: ogrid_nx_bak, ogrid_ny_bak, ogrid_nz_bak
    if (ogrid%ax%n/=ogrid_nx_bak .or. ogrid%ay%n/=ogrid_ny_bak .or. ogrid%az%n/=ogrid_nz_bak) then
       select case (pgrid%mode)
         case ("imgfwd")
              if (igrid%cpximg) then
                 deallocate(cimg)
                 allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              else
                 deallocate(rimg)
                 allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              end if
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("imgadj")
              if (igrid%cpximg) then
                 deallocate(cimg)
                 allocate(cimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              else
                 deallocate(rimg)
                 allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              end if
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("tomfwd")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("tomadj")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("tominv")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              deallocate(index_mxr,index_myr)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))

         case ("hesian")
              deallocate(rimg)
              deallocate(index_mxs,index_mys)
              allocate(rimg(ogrid%ahx%n,ogrid%ahy%n,ogrid%ax%n,ogrid%ay%n,nth))
              allocate(index_mxs(ogrid%ahx%n,ogrid%ax%n),index_mys(ogrid%ahy%n,ogrid%ay%n))
              if (hgrid%hestyp==2) then
                 deallocate(index_mxr,index_myr)
                  allocate(index_mxr(ogrid%ahx%n,ogrid%ax%n),index_myr(ogrid%ahy%n,ogrid%ay%n))
              end if
       end select
    end if
    return
  end subroutine

  subroutine bwi_module_init()
    integer :: stat

    call k2_map(kx2, pgrid%ax%n, pgrid%ax%d) 
    call k2_map(ky2, pgrid%ay%n, pgrid%ay%d)

    stat = bwi_tapers_init(tapx, tapy, pgrid%ntpx, pgrid%ntpy)

    select case (pgrid%mode)

      case ("imgfwd")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("imgadj")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("tomfwd")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2dtF, plan_2dtI, wtmp_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("tomadj")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2dtF, plan_2dtI, wtmp_slice, pgrid%ax%n, pgrid%ay%n, nth)

      case ("tominv")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fftw_init(plan_2dtF, plan_2dtI, wtmp_slice, pgrid%ax%n, pgrid%ay%n, nth)
           call bwi_fio_dvel_init(vgrid, ftag_%grdo_)
           call bwi_fio_dvel_init(vgrid, ftag_%grdp_)
           call bwi_fio_dvel_init(vgrid, ftag_%srch_)

      case ("hesian")
           call bwi_fftw_init(plan_2dsF, plan_2dsI, dwfs_slice, pgrid%ax%n, pgrid%ay%n, nth)
           if (hgrid%hestyp==2) then
              call bwi_fftw_init(plan_2drF, plan_2drI, dwfr_slice, pgrid%ax%n, pgrid%ay%n, nth)
           end if

    end select
    return
  end subroutine

end module
